@@ -4,6 +4,12 @@
 
 struct tm;              /* declaration */
 
+/**
+ * max size of a buffer which will just be reset
+ * to ->used = 0 instead of really freeing the buffer
+ */
+#define BUFFER_MAX_REUSE_SIZE 4096
+
 /* generic string + binary data container; contains a terminating 0 in both
  * cases
  *
@@ -19,28 +25,32 @@ typedef struct {
 	char *ptr;
 
 	/* "used" includes a terminating 0 */
-	size_t used;
+	uint32_t used;
 	/* size of allocated buffer at *ptr */
-	size_t size;
+	uint32_t size;
 } buffer;
 
 /* create new buffer; either empty or copy given data */
+__attribute_returns_nonnull__
 buffer* buffer_init(void);
+
+__attribute_returns_nonnull__
 buffer* buffer_init_buffer(const buffer *src); /* src can  be NULL */
+
+__attribute_returns_nonnull__
 buffer* buffer_init_string(const char *str); /* str can  be NULL */
 
 void buffer_free(buffer *b); /* b can be NULL */
-/* truncates to used == 0; frees large buffers, might keep smaller ones for reuse */
-void buffer_reset(buffer *b); /* b can be NULL */
 
 /* reset b. if NULL != b && NULL != src, move src content to b. reset src. */
-void buffer_move(buffer *b, buffer *src);
+void buffer_move(buffer * restrict b, buffer * restrict src);
 
 /* make sure buffer is large enough to store a string of given size
  * and a terminating zero.
  * sets b to an empty string, and may drop old content.
  * @return b->ptr
  */
+__attribute_returns_nonnull__
 char* buffer_string_prepare_copy(buffer *b, size_t size);
 
 /* allocate buffer large enough to be able to append a string of given size
@@ -49,6 +59,7 @@ char* buffer_string_prepare_copy(buffer *b, size_t size);
  * "used" data is preserved; if not empty buffer must contain a
  * zero terminated string.
  */
+__attribute_returns_nonnull__
 char* buffer_string_prepare_append(buffer *b, size_t size);
 
 /* use after prepare_(copy,append) when you have written data to the buffer
@@ -63,7 +74,7 @@ void buffer_commit(buffer *b, size_t size);
  * - does not modify the string data apart from terminating zero
  * - reallocates the buffer iff needed
  */
-void buffer_string_set_length(buffer *b, size_t len);
+void buffer_string_set_length(buffer *b, uint32_t len);
 
 /* clear buffer
  * - invalidate buffer contents
@@ -72,31 +83,45 @@ void buffer_string_set_length(buffer *b, size_t len);
  */
 static inline void buffer_clear(buffer *b);
 
-void buffer_copy_string(buffer *b, const char *s);
-void buffer_copy_string_len(buffer *b, const char *s, size_t s_len);
-static inline void buffer_copy_buffer(buffer *b, const buffer *src);
+/* reset buffer
+ * - invalidate buffer contents
+ * - unsets used chars
+ * - keeps smaller buffer (unmodified) for reuse
+ *   (b->ptr *is not* set to an empty, '\0'-terminated string "")
+ * - frees larger buffer (b->size > BUFFER_MAX_REUSE_SIZE)
+ */
+static inline void buffer_reset(buffer *b);
 
-void buffer_append_string(buffer *b, const char *s);
-void buffer_append_string_len(buffer *b, const char *s, size_t s_len);
-static inline void buffer_append_string_buffer(buffer *b, const buffer *src);
+/* free buffer ptr
+ * - invalidate buffer contents; free ptr; reset ptr, used, size to 0
+ */
+__attribute_cold__
+void buffer_free_ptr(buffer *b);
+
+void buffer_copy_string(buffer * restrict b, const char * restrict s);
+void buffer_copy_string_len(buffer * restrict b, const char * restrict s, size_t s_len);
+static inline void buffer_copy_buffer(buffer * restrict b, const buffer * restrict src);
+
+void buffer_append_string(buffer * restrict b, const char * restrict s);
+void buffer_append_string_len(buffer * restrict b, const char * restrict s, size_t s_len);
+static inline void buffer_append_string_buffer(buffer * restrict b, const buffer * restrict src);
 
 #define buffer_append_uint_hex(b,len) buffer_append_uint_hex_lc((b),(len))
 void buffer_append_uint_hex_lc(buffer *b, uintmax_t len);
 void buffer_append_int(buffer *b, intmax_t val);
-void buffer_copy_int(buffer *b, intmax_t val);
 
-void buffer_append_strftime(buffer *b, const char *format, const struct tm *tm);
+void buffer_append_strftime(buffer * restrict b, const char * restrict format, const struct tm * restrict tm);
 
 /* '-', log_10 (2^bits) = bits * log 2 / log 10 < bits * 0.31, terminating 0 */
 #define LI_ITOSTRING_LENGTH (2 + (8 * sizeof(intmax_t) * 31 + 99) / 100)
 
-void li_itostrn(char *buf, size_t buf_len, intmax_t val);
-void li_utostrn(char *buf, size_t buf_len, uintmax_t val);
+size_t li_itostrn(char *buf, size_t buf_len, intmax_t val);
+size_t li_utostrn(char *buf, size_t buf_len, uintmax_t val);
 
 /* buf must be (at least) 2*s_len + 1 big. uses lower-case hex letters. */
 #define li_tohex(buf,buf_len,s,s_len) li_tohex_lc((buf),(buf_len),(s),(s_len))
-void li_tohex_lc(char *buf, size_t buf_len, const char *s, size_t s_len);
-void li_tohex_uc(char *buf, size_t buf_len, const char *s, size_t s_len);
+void li_tohex_lc(char * restrict buf, size_t buf_len, const char * restrict s, size_t s_len);
+void li_tohex_uc(char * restrict buf, size_t buf_len, const char * restrict s, size_t s_len);
 
 /* NULL buffer or empty buffer (used == 0);
  * unset "string" (buffer) config options are initialized to used == 0,
@@ -132,10 +157,10 @@ int buffer_is_equal_string(const buffer *a, const char *s, size_t b_len);
 __attribute_pure__
 int buffer_is_equal_caseless_string(const buffer *a, const char *s, size_t b_len);
 
-void buffer_substr_replace (buffer *b, size_t offset, size_t len, const buffer *replace);
+void buffer_substr_replace (buffer * restrict b, size_t offset, size_t len, const buffer * restrict replace);
 
-void buffer_append_string_encoded_hex_lc(buffer *b, const char *s, size_t len);
-void buffer_append_string_encoded_hex_uc(buffer *b, const char *s, size_t len);
+void buffer_append_string_encoded_hex_lc(buffer * restrict b, const char * restrict s, size_t len);
+void buffer_append_string_encoded_hex_uc(buffer * restrict b, const char * restrict s, size_t len);
 
 typedef enum {
 	ENCODING_REL_URI, /* for coding a rel-uri (/with space/and%percent) nicely as part of a href */
@@ -144,13 +169,13 @@ typedef enum {
 	ENCODING_MINIMAL_XML   /* minimal encoding for xml */
 } buffer_encoding_t;
 
-void buffer_append_string_encoded(buffer *b, const char *s, size_t s_len, buffer_encoding_t encoding);
+void buffer_append_string_encoded(buffer * restrict b, const char * restrict s, size_t s_len, buffer_encoding_t encoding);
 
 /* escape non-printable characters; simple escapes for \t, \r, \n; fallback to \xCC */
-void buffer_append_string_c_escaped(buffer *b, const char *s, size_t s_len);
+void buffer_append_string_c_escaped(buffer * restrict b, const char * restrict s, size_t s_len);
 
 /* to upper case, replace non alpha-numerics with '_'; if is_http_header prefix with "HTTP_" unless s is "content-type" */
-void buffer_copy_string_encoded_cgi_varnames(buffer *b, const char *s, size_t s_len, int is_http_header);
+void buffer_copy_string_encoded_cgi_varnames(buffer * restrict b, const char * restrict s, size_t s_len, int is_http_header);
 
 void buffer_urldecode_path(buffer *url);
 void buffer_urldecode_query(buffer *url);
@@ -168,19 +193,19 @@ char int2hex(char i);
 __attribute_pure__
 static inline int light_isdigit(int c);
 static inline int light_isdigit(int c) {
-	return (c >= '0' && c <= '9');
+	return ((uint32_t)c-'0' <= '9'-'0');
 }
 
 __attribute_pure__
 static inline int light_isxdigit(int c);
 static inline int light_isxdigit(int c) {
-	return light_isdigit(c) || (c |= 32, c >= 'a' && c <= 'f');
+	return light_isdigit(c) || (((uint32_t)c | 0x20)-'a' <= 'f'-'a');
 }
 
 __attribute_pure__
 static inline int light_isalpha(int c);
 static inline int light_isalpha(int c) {
-	return (c |= 32, c >= 'a' && c <= 'z');
+	return (((uint32_t)c | 0x20)-'a' <= 'z'-'a');
 }
 
 __attribute_pure__
@@ -189,15 +214,23 @@ static inline int light_isalnum(int c) {
 	return light_isdigit(c) || light_isalpha(c);
 }
 
+#define light_isupper(c) ((uint32_t)(c)-'A' <= 'Z'-'A')
+#define light_islower(c) ((uint32_t)(c)-'a' <= 'z'-'a')
+
+#define light_bshift(b)           ((uint64_t)1uL << (b))
+#define light_btst(a,b)  ((a) &   ((uint64_t)1uL << (b)))
+#define light_bclr(a,b)  ((a) &= ~((uint64_t)1uL << (b)))
+#define light_bset(a,b)  ((a) |=  ((uint64_t)1uL << (b)))
+
 
 __attribute_pure__
-static inline size_t buffer_string_length(const buffer *b); /* buffer string length without terminating 0 */
+static inline uint32_t buffer_string_length(const buffer *b); /* buffer string length without terminating 0 */
 
 __attribute_pure__
-static inline size_t buffer_string_space(const buffer *b); /* maximum length of string that can be stored without reallocating */
+static inline uint32_t buffer_string_space(const buffer *b); /* maximum length of string that can be stored without reallocating */
 
 static inline void buffer_append_slash(buffer *b); /* append '/' no non-empty strings not ending in '/' */
-void buffer_append_path_len(buffer *b, const char *a, size_t alen); /* join strings with '/', if '/' not present */
+void buffer_append_path_len(buffer * restrict b, const char * restrict a, size_t alen); /* join strings with '/', if '/' not present */
 
 #define BUFFER_APPEND_STRING_CONST(x, y) \
 	buffer_append_string_len(x, y, sizeof(y) - 1)
@@ -207,7 +240,8 @@ void buffer_append_path_len(buffer *b, const char *a, size_t alen); /* join stri
 
 #define BUFFER_INTLEN_PTR(x) (x)->used ? (int)((x)->used - 1) : 0, (x)->ptr
 
-#define CONST_STR_LEN(x) x, (x) ? sizeof(x) - 1 : 0
+#define CONST_LEN_STR(x) (uint32_t)sizeof(x)-1, x
+#define CONST_STR_LEN(x) x, (uint32_t)sizeof(x) - 1
 #define CONST_BUF_LEN(x) ((x) ? (x)->ptr : NULL), buffer_string_length(x)
 
 
@@ -227,24 +261,24 @@ static inline int buffer_string_is_empty(const buffer *b) {
 	return NULL == b || b->used < 2;
 }
 
-static inline size_t buffer_string_length(const buffer *b) {
+static inline uint32_t buffer_string_length(const buffer *b) {
 	return NULL != b && 0 != b->used ? b->used - 1 : 0;
 }
 
-static inline size_t buffer_string_space(const buffer *b) {
+static inline uint32_t buffer_string_space(const buffer *b) {
 	return NULL != b && b->size ? b->size - (b->used | (0 == b->used)) : 0;
 }
 
-static inline void buffer_copy_buffer(buffer *b, const buffer *src) {
+static inline void buffer_copy_buffer(buffer * restrict b, const buffer * restrict src) {
 	buffer_copy_string_len(b, CONST_BUF_LEN(src));
 }
 
-static inline void buffer_append_string_buffer(buffer *b, const buffer *src) {
+static inline void buffer_append_string_buffer(buffer * restrict b, const buffer * restrict src) {
 	buffer_append_string_len(b, CONST_BUF_LEN(src));
 }
 
 static inline void buffer_append_slash(buffer *b) {
-	size_t len = buffer_string_length(b);
+	uint32_t len = buffer_string_length(b);
 	if (len > 0 && '/' != b->ptr[len-1]) BUFFER_APPEND_STRING_CONST(b, "/");
 }
 
@@ -252,4 +286,11 @@ static inline void buffer_clear(buffer *b) {
 	b->used = 0;
 }
 
+static inline void buffer_reset(buffer *b) {
+	b->used = 0;
+	/* release buffer larger than BUFFER_MAX_REUSE_SIZE bytes */
+	if (b->size > BUFFER_MAX_REUSE_SIZE) buffer_free_ptr(b);
+}
+
+
 #endif
