@@ -2,8 +2,8 @@
 
 #include "base.h"
 #include "burl.h"
-#include "etag.h"
 #include "fdevent.h"
+#include "http_etag.h"
 #include "keyvalue.h"
 #include "log.h"
 #include "stream.h"
@@ -15,7 +15,9 @@
 #include "sys-crypto.h"
 
 #include <sys/stat.h>
+#ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
+#endif
 
 #include <stdlib.h>
 #include <fcntl.h>
@@ -2003,7 +2005,7 @@ int config_parse_file(server *srv, config_t *context, const char *fn) {
 		filename = buffer_init_string(fn);
 	} else {
 		filename = buffer_init_buffer(context->basedir);
-		buffer_append_string(filename, fn);
+		buffer_append_path_len(filename, fn, strlen(fn));
 	}
 
 	switch (glob(filename->ptr, flags, NULL, &gl)) {
@@ -2087,7 +2089,6 @@ int config_parse_cmd(server *srv, config_t *context, const char *cmd) {
 		}
 		else {
 			ssize_t rd;
-			pid_t wpid;
 			int wstatus = 0;
 			buffer *out = buffer_init();
 			close(fds[1]);
@@ -2102,8 +2103,7 @@ int config_parse_cmd(server *srv, config_t *context, const char *cmd) {
 			}
 			close(fds[0]);
 			fds[0] = -1;
-			while (-1 == (wpid = waitpid(pid, &wstatus, 0)) && errno == EINTR) ;
-			if (wpid != pid) {
+			if (pid != fdevent_waitpid(pid, &wstatus, 0)) {
 				log_perror(srv->errh, __FILE__, __LINE__, "waitpid \"%s\"",cmd);
 				ret = -1;
 			}
@@ -2257,14 +2257,17 @@ int config_set_defaults(server *srv) {
 		buffer_clear(b);
 		if (!buffer_string_is_empty(srv->srvconf.changeroot)) {
 			buffer_copy_buffer(b, srv->srvconf.changeroot);
-			buffer_append_slash(b);
 		}
 		len = buffer_string_length(b);
 
 		for (i = 0; i < srv->srvconf.upload_tempdirs->used; ++i) {
 			const data_string * const ds = (data_string *)srv->srvconf.upload_tempdirs->data[i];
-			buffer_string_set_length(b, len); /*(truncate)*/
-			buffer_append_string_buffer(b, &ds->value);
+			if (len) {
+				buffer_string_set_length(b, len); /*(truncate)*/
+				buffer_append_path_len(b, CONST_BUF_LEN(&ds->value));
+			} else {
+				buffer_copy_buffer(b, &ds->value);
+			}
 			if (-1 == stat(b->ptr, &st1)) {
 				log_error(srv->errh, __FILE__, __LINE__,
 				  "server.upload-dirs doesn't exist: %s", b->ptr);
